"use strict";
var _errors_1 = require("./_errors");
/**
 * Allows checking of test outcomes
 * @param actualValue - the value or function under test
 */
function Expect(actualValue) {
    return new Matcher(actualValue);
}
exports.Expect = Expect;
/**
 * Gives functionality to ensure the outcome of a test is as expected
 */
var Matcher = (function () {
    function Matcher(actualValue) {
        this._shouldMatch = true;
        this._actualValue = actualValue;
    }
    Object.defineProperty(Matcher.prototype, "not", {
        /**
         * Any subsequent matcher function will be looking for the opposite criteria
         */
        get: function () {
            this._shouldMatch = !this._shouldMatch;
            return this;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Checks that a value is identical to another
     * @param expectedValue - the value that will be used to match
     */
    Matcher.prototype.toBe = function (expectedValue) {
        if (expectedValue !== this._actualValue === this._shouldMatch) {
            throw new _errors_1.ExactMatchError(this._actualValue, expectedValue, this._shouldMatch);
        }
    };
    /**
     * Checks that a value is equal to another (for objects the function will check for deep equality)
     * @param expectedValue - the value that will be used to match
     */
    Matcher.prototype.toEqual = function (expectedValue) {
        // exclude the double equals in this case from review as this is what we want to do
        if (expectedValue != this._actualValue === this._shouldMatch) {
            if (typeof expectedValue !== "object" || this._checkObjectsAreDeepEqual(expectedValue, this._actualValue) !== this._shouldMatch) {
                throw new _errors_1.EqualMatchError(this._actualValue, expectedValue, this._shouldMatch);
            }
        }
    };
    Matcher.prototype._checkObjectsAreDeepEqual = function (objectA, objectB) {
        // if one object is an array and the other is not then they are not equal
        if (Array.isArray(objectA) !== Array.isArray(objectB)) {
            return false;
        }
        // get all the property keys for each object
        var objectAKeys = Object.keys(objectA);
        var objectBKeys = Object.keys(objectB);
        // if they don't have the same amount of properties then clearly not
        if (objectAKeys.length !== objectBKeys.length) {
            return false;
        }
        // check all the properties of each object
        for (var i = 0; i < objectAKeys.length; i++) {
            var objectAKey = objectAKeys[i];
            // if the property values are not the same
            if (objectA[objectAKey] !== objectB[objectAKey]) {
                // and it's not an object or the objects are not equal
                if (typeof (objectA[objectAKey]) !== "object" || this._checkObjectsAreDeepEqual(objectA[objectAKey], objectB[objectAKey]) === false) {
                    // then not deep equal
                    return false;
                }
            }
        }
        // all properties match so all is good
        return true;
    };
    /**
     * Checks that a value conforms to a regular expression
     * @param regex - the regular expression that the actual value should match
     */
    Matcher.prototype.toMatch = function (regex) {
        if (!regex.test(this._actualValue) === this._shouldMatch) {
            throw new _errors_1.RegexMatchError(this._actualValue, regex, this._shouldMatch);
        }
    };
    /**
     * Checks that a value is not undefined
     */
    Matcher.prototype.toBeDefined = function () {
        if (this._actualValue === undefined === this._shouldMatch) {
            throw new _errors_1.ExactMatchError(this._actualValue, undefined, !this._shouldMatch);
        }
    };
    /**
     * Checks that a value is null
     */
    Matcher.prototype.toBeNull = function () {
        if (this._actualValue !== null === this._shouldMatch) {
            throw new _errors_1.ExactMatchError(this._actualValue, null, this._shouldMatch);
        }
    };
    /**
     * Checks that a value is equivalent to boolean true
     */
    Matcher.prototype.toBeTruthy = function () {
        if ((this._actualValue && !this._shouldMatch) || (!this._actualValue && this._shouldMatch)) {
            throw new _errors_1.TruthyMatchError(this._actualValue, this._shouldMatch);
        }
    };
    /**
     * Checks that a string contains another string or an array contains a specific item
     * @param expectedContent - the string or array item that the value should contain
     */
    Matcher.prototype.toContain = function (expectedContent) {
        if (this._actualValue.indexOf(expectedContent) === -1 === this._shouldMatch) {
            throw new _errors_1.ContentsMatchError(this._actualValue, expectedContent, this._shouldMatch);
        }
    };
    /**
     * Checks that a number is less than a given limit
     * @param upperLimit - the number that the number under test should be less than
     */
    Matcher.prototype.toBeLessThan = function (upperLimit) {
        if (this._actualValue < upperLimit !== this._shouldMatch) {
            throw new _errors_1.LessThanMatchError(this._actualValue, upperLimit, this._shouldMatch);
        }
    };
    /**
     * Checks that a number is greater than a given limit
     * @param lowerLimit - the number that the number under test should be greater than
     */
    Matcher.prototype.toBeGreaterThan = function (lowerLimit) {
        if (this._actualValue > lowerLimit !== this._shouldMatch) {
            throw new _errors_1.GreaterThanMatchError(this._actualValue, lowerLimit, this._shouldMatch);
        }
    };
    /**
     * Checks that a function throws an error when executed
     */
    Matcher.prototype.toThrow = function () {
        var errorThrown;
        try {
            this._actualValue();
        }
        catch (error) {
            if (!this._shouldMatch) {
                throw new _errors_1.ErrorMatchError(error, this._shouldMatch);
            }
            errorThrown = error;
        }
        if (this._shouldMatch && errorThrown === undefined) {
            throw new _errors_1.ErrorMatchError(undefined, this._shouldMatch);
        }
    };
    /**
     * Checks that a function throws a specific error
     * @param errorType - the type of error that should be thrown
     * @param errorMessage - the message that the error should have
     */
    Matcher.prototype.toThrowError = function (errorType, errorMessage) {
        var threwRightError = false;
        var actualError;
        try {
            this._actualValue();
        }
        catch (error) {
            actualError = error;
            if (error instanceof errorType && error.message === errorMessage) {
                threwRightError = true;
            }
        }
        if (!threwRightError === this._shouldMatch) {
            throw new _errors_1.ErrorMatchError(actualError, this._shouldMatch, errorType, errorMessage);
        }
    };
    /**
     * Checks that a spy has been called
     */
    Matcher.prototype.toHaveBeenCalled = function () {
        if (this._actualValue.calls.length === 0 === this._shouldMatch) {
            throw new _errors_1.FunctionCallMatchError(this._actualValue, this._shouldMatch);
        }
    };
    /**
     * Checks that a spy has been called with the specified arguments
     * @param args - a list of arguments that the spy should have been called with
     */
    Matcher.prototype.toHaveBeenCalledWith = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i - 0] = arguments[_i];
        }
        if (this._actualValue.calls.filter(function (call) {
            return call.args.filter(function (arg, index) { return arg === args[index]; }).length === args.length &&
                call.args.length === args.length; // and the call has the same amount of arguments
        }).length === 0 === this._shouldMatch) {
            throw new _errors_1.FunctionCallMatchError(this._actualValue, this._shouldMatch, args);
        }
    };
    /**
     * Checks that a property spy has been set
     */
    Matcher.prototype.toHaveBeenSet = function () {
        if (this._actualValue.setCalls.length === 0 === this._shouldMatch) {
            throw new _errors_1.PropertySetMatchError(this._actualValue, this._shouldMatch);
        }
    };
    /**
     * Checks that a property spy has been set to a specific value
     * @param value - a value to which the property should be set to
     */
    Matcher.prototype.toHaveBeenSetTo = function (value) {
        if (this._actualValue.setCalls.filter(function (call) { return call.args[0] === value; }).length === 0 === this._shouldMatch) {
            throw new _errors_1.PropertySetMatchError(this._actualValue, this._shouldMatch, value);
        }
    };
    return Matcher;
}());
exports.Matcher = Matcher;
//# sourceMappingURL=expect.js.map